-module(eTpf).

-include("eTpf.hrl").

-export([
	%% 记录函数
	sts/0             %% tpTracerShell
	, stl/0           %% tpTracerLog
	, stf/0           %% tpTracerFile

	%% pt函数
	, pts/0           %% tpTracerShell
	, ptl/0           %% tpTracerLog
	, ptf/0           %% tpTracerFile

	%% pf函数
	, rsPfs/1
	, rsPfm/2

	, sdPfs/2
	, sdPfm/3

	, flPfs/2
	, flPfm/3

	, cgPfs/2
	, cgPfs/3
	, cgPfm/3
	, cgPfm/4

	, gcPfs/1
	, gcPfm/2

	%% trace函数
	, trace/1
	, trace/2
	, trace/3
	, trace/4
	, stop/0
	, stop/1
]).

-compile([export_all]).

sts() ->
	trace([{app, eSync}, {scope, [whereis(esSyncSrv)]}]).

sts1() ->
	trace([test, {scope, [self()]}]).

stl() ->
	trace([test, {scope, [self()]}], tpTracerLog).

stf() ->
	trace([{app, eSync}, {scope, [whereis(esSyncSrv)]}], tpTracerFile).

stop() ->
	stop(?eTpfTracerId).

stop(TracerId) ->
	supervisor:terminate_child(eTpf_sup, TracerId).

pts() ->
	ok.

ptl() ->
	ok.


ptf() ->
	ok.

rsPfs(InputFile) ->
	tpMsgRS:pfs(InputFile).

rsPfm(InputFiles, Cwd) ->
	tpMsgRS:pfm(InputFiles, Cwd).

sdPfs(InputFile, Pids) ->
	tpMsgSD:pfs(InputFile, Pids).

sdPfm(InputFiles, Cwd, Pids) ->
	tpMsgSD:pfm(InputFiles, Cwd, Pids).

flPfs(InputFile, OutputPath) ->
	tpFlame:pfs(InputFile, OutputPath).

flPfm(InputFiles, Cwd, OutputPath) ->
	tpFlame:pfm(InputFiles, Cwd, OutputPath).

cgPfs(InputFile, Output) ->
	tpCallGrind:pfs(InputFile, Output).

cgPfs(InputFile, Output, Opts) ->
	tpCallGrind:pfs(InputFile, Output, Opts).

cgPfm(Wildcard, Cwd, Prefix) ->
	tpCallGrind:pfm(Wildcard, Cwd, Prefix).

cgPfm(Wildcard, Cwd, Prefix, Opts) ->
	tpCallGrind:pfm(Wildcard, Cwd, Prefix, Opts).

gcPfs(InputFile) ->
	tpGcColl:pfs(InputFile).

gcPfm(InputFiles, Cwd) ->
	tpGcColl:pfm(InputFiles, Cwd).

-spec trace(userInput()) -> ok.
trace(Input) ->
	trace(Input, tpTracerShell).

-spec trace(userInput(), module()) -> ok.
trace(Input, TracerMod) ->
	trace(Input, TracerMod, #{}, #{}).

-spec trace(userInput(), module(), tracerOpts()) -> ok.
trace(Input, TracerMod, TracerOpts) ->
	trace(Input, TracerMod, TracerOpts, #{}).

-spec trace(userInput(), module(), tracerOpts(), traceOpts()) -> ok.
trace(Input, TracerMod, TracerOpts, TraceOpts) when is_list(Input) ->
	InputList = case is_list(Input) of true -> Input; _ -> [Input] end,
	doTrace(InputList, TracerMod, TracerOpts, TraceOpts).

doTrace(PidPortSpec, FlagList, TraceMods, TracerMod, TracerOpts, TraceOpts) ->
	_ = application:ensure_all_started(eTpf),

	TracerId = maps:get(tracerId, TraceOpts, ?eTpfTracerId),
	case supervisor:start_child(eTpf_sup, #{
		id => TracerId,
		start => {TracerMod, start_link, [TracerOpts]},
		restart => temporary,
		type => worker,
		shutdown => infinity
	}) of
		{ok, TracerPid} ->
			_Mode = maps:get(mode, TraceOpts, trace),
			LastTraceMods = flattenMods(TraceMods, []),
			traceInput(PidPortSpec, FlagList, LastTraceMods, TracerPid, TraceOpts),
			ok;
		_Err ->
			io:format("trace start error ~p~n", [_Err])
	end.

flattenMods([], Acc) ->
	case lists:flatten(Acc) of
		[] ->
			['_'];
		AllMods ->
			AllMods
	end;
flattenMods([{callback, Mod, Fun} | Tail], Acc) when is_atom(Mod), is_atom(Fun) ->
	Input = flattenMods(Mod:Fun(), []),
	flattenMods(Tail, [Input | Acc]);
flattenMods([{app, App} | Tail], Acc) when is_atom(App) ->
	_ = application:load(App),
	{ok, Mods} = application:get_key(App, modules),
	flattenMods(Tail, [Mods | Acc]);
flattenMods([Input | Tail], Acc) ->
	flattenMods(Tail, [Input | Acc]).

traceInput(PidPortSpec, FlagList, TraceMods, TracerState, TraceOpts) ->
	% 每个进程只能由一个tracer进行跟踪。因此，跟踪已跟踪进程会失败。
	%% @todo It might be useful to count the number of sends
	%% or receives a function does.
	%ExtraFlags = [running || maps:get(running, TraceOpts, false)] ++ [send || maps:get(send, TraceOpts, false)],
	[
		begin
		%erlang:trace(PidPortSpec, true, [call, procs, arity, return_to, set_on_spawn, {tracer, tpTracerNif, TracerState} | ExtraFlags])
			erlang:trace(OnePidPortSpec, true, [{tracer, tpTracerNif, TracerState} | FlagList])
		end || OnePidPortSpec <- PidPortSpec
	],

	MatchSpec =
		case TraceOpts of
			#{process_dump := true} ->
				[{'_', [], [{message, {process_dump}}]}];
			_ ->
				true
		end,

	[
	   begin
		%% The module must be loaded before we attempt to trace it.
			_ = code:ensure_loaded(OneTraceMods),
			io:format("IMY******************  ~p~n  ", [OneTraceMods]),
			_ = erlang:trace_pattern({OneTraceMods, '_', '_'}, [{'_', [], [{message, {process_dump}}]}], [local])
		end || OneTraceMods <- TraceMods
	],
	ok.
